#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import json, os

from netaddr import *
from oslo_utils import netutils
from sysrepo_api import sysrepo_api3 as api


class SysrepoDhcpServer(api.SysrepoAPI):

    def __init__(self,
                 app_name="dhcpserver",
                 log_file=None,
                 log_level=None,
                 console_log=False,
                 namespace="/ucpe:config/dhcpserver:dhcp-server",
                 run_as_subscriber=False,
                 priority=0):
        api.SysrepoAPI.__init__(
            self,
            app_name=app_name,
            log_file=log_file,
            log_level=log_level,
            console_log=console_log,
            namespace=namespace,
            run_as_subscriber=run_as_subscriber,
            priority=priority)

    def isIPv6(self, ipaddr):
        """Check the IP address version and return True in case of IPv6"""
        if ipaddr is not None and ipaddr != "":
            str_rng = ipaddr.rfind('-')
            return netutils.is_valid_ipv6(ipaddr[:str_rng])

        return False

    def get_system_bind_iface_address(self, iface=""):
        address = self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv4-prefix".format(iface))
        if address != '' and address is not None:
            ip = IPNetwork(address)
            return str(ip.ip)
        else:
            return None

    def get_system_bind_iface6_address(self, iface=""):
        address = self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv6-prefix".format(iface))
        if address != '' and address is not None:
            ip = IPNetwork(address)
            return str(ip.ip)
        else:
            return None

    def get_system_bind_iface_netmask(self, iface=""):
        address = self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv4-prefix".format(iface))
        if address != '' and address is not None:
            ip = IPNetwork(address)
            return str(ip.netmask)
        else:
            return None

    def get_system_bind_iface6_netmask(self, iface=""):
        address = self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv6-prefix".format(iface))
        if address != '' and address is not None:
            ip = IPNetwork(address)
            return str(ip.prefixlen)
        else:
            return None

    def apply_dhcpserver2system(self, addr, netmask, dhcp_range, router, dns, dlt, mlt):
        if addr is None:
            return False
        if netmask is None:
            return False
        if dhcp_range is None:
            return False
        ind = addr.rfind('.')
        fd = open('/etc/dhcp/dhcpd.conf', 'w')
        if dlt:
            fd.write("default-lease-time {};\n".format(dlt))

        if mlt:
            fd.write("max-lease-time {};\n".format(mlt))

        fd.write("subnet " + addr[:ind + 1] + "0 netmask " + netmask + " {\n")
        if router:
            fd.write("    option routers {};\n".format(router))
        if dns:
            fd.write("    option domain-name-servers {};\n".format(dns))

        fd.write("    range " + (dhcp_range.replace("-", " ")) + ";\n}\n")
        fd.flush()
        os.fsync(fd.fileno())
        fd.close()

        return True

    def apply_dhcp6server2system(self, addr, netmask, dhcp_range, router, dns, dlt, mlt, dhcp_int):
        if addr is None:
            self.logger.error('ERROR: Missing network address !')
            return False
        if netmask is None:
            self.logger.error('ERROR: Missing network netmask !')
            return False
        if dhcp_range is None:
            self.logger.error('ERROR: Missing IPv6 range !')
            return False
        ind = addr.rfind('::')
        fd = open('/etc/dhcp/dhcpd6.conf', 'w')
        fd.write("preferred-lifetime 86400;\n")
        if dlt:
            fd.write("default-lease-time {};\n".format(dlt))
        else:
            fd.write("default-lease-time 86400;\n")

        if mlt:
            fd.write("max-lease-time {};\n".format(mlt))
        else:
            fd.write("max-lease-time 172800;\n")

        fd.write("ddns-updates on;\nddns-update-style interim;\nallow client-updates;\nauthoritative;\n")
        fd.write("subnet6 " + addr[:ind + 2] + "/" + netmask + " {\n")

        if dns:
            fd.write("    option domain-name-servers {};\n".format(dns))

        fd.write("    range6 " + (dhcp_range.replace("-", " ")) + ";\n}\n")
        fd.flush()
        os.fsync(fd.fileno())
        fd.close()

        if dhcp_int:
            with open('/etc/default/isc-dhcp-server', 'w') as f:
                f.write('INTERFACESv6="br-{}"\n'.format(dhcp_int))

        return True

    def start_dhcp_service(self, service):
        if service == "isc-dhcp-server.service":
            self.stop_dhcp_service("isc-dhcp-server6.service")
        else:
            self.stop_dhcp_service("isc-dhcp-server.service")

        self.execute_shell_command("systemctl enable {}".format(service))
        self.execute_shell_command("systemctl --no-block restart {}".format(service))

    def stop_dhcp_service(self, service):
        self.execute_shell_command("systemctl --no-block stop {}".format(service))
        self.execute_shell_command("systemctl reset-failed {} || true".format(service))
        self.execute_shell_command("systemctl disable {}".format(service))

    def apply_startup(self):
        self.logger.debug('Apply STARTUP config for DHCP Server')
        sys_conf = self.get_startup_config(full_config=False, full_path=False)

        self.logger.debug('Startup config : %s' % json.dumps(sys_conf))
        dhcp_state = False
        dhcp_range = ''
        dhcp_int = ''
        router = None
        dns = None
        dlt = None
        mlt = None

        if 'enabled' in sys_conf:
            dhcp_state = sys_conf['enabled']
        if 'dhcp-range' in sys_conf:
            dhcp_range = sys_conf['dhcp-range']
        if 'dhcp-bind-interface' in sys_conf:
            dhcp_int = sys_conf['dhcp-bind-interface']
        if 'router' in sys_conf:
            router = sys_conf['router']
        if 'domain-name-servers' in sys_conf:
            dns = sys_conf['domain-name-servers']
        if 'default-lease-time' in sys_conf:
            dlt = sys_conf['default-lease-time']
        if 'max-lease-time' in sys_conf:
            mlt = sys_conf['max-lease-time']

        if dhcp_state is True:
            if self.isIPv6(dhcp_range):
                dhcp_service = "isc-dhcp-server6.service"
                addr = self.get_system_bind_iface6_address(dhcp_int)
                netmask = self.get_system_bind_iface6_netmask(dhcp_int)
                res = self.apply_dhcp6server2system(addr, netmask, dhcp_range, router, dns, dlt, mlt,dhcp_int)
            else:
                dhcp_service = "isc-dhcp-server.service"
                addr = self.get_system_bind_iface_address(dhcp_int)
                netmask = self.get_system_bind_iface_netmask(dhcp_int)
                res = self.apply_dhcpserver2system(addr, netmask, dhcp_range, router, dns, dlt, mlt)

            if res is False:
                self.logger.error('Failed creating DHCP server configuration file')
                return True
            self.start_dhcp_service(dhcp_service)
        else:
            self.stop_dhcp_service("isc-dhcp-server.service")
            self.stop_dhcp_service("isc-dhcp-server6.service")

        return True

    def apply_running(self, change_set):
        self.logger.debug('Apply RUNNING config for DHCP Server')
        json_data = self.get_running_config(full_path=False)
        dhcp_state = False
        dhcp_range = ''
        dhcp_int = ''
        router = None
        dns = None
        dlt = None
        mlt = None

        if 'enabled' in json_data:
            dhcp_state = json_data['enabled']
        if 'dhcp-range' in json_data:
            dhcp_range = json_data['dhcp-range']
        if 'dhcp-bind-interface' in json_data:
            dhcp_int = json_data['dhcp-bind-interface']
        if 'router' in json_data:
            router = json_data['router']
        if 'domain-name-servers' in json_data:
            dns = json_data['domain-name-servers']
        if 'default-lease-time' in json_data:
            dlt = json_data['default-lease-time']
        if 'max-lease-time' in json_data:
            mlt = json_data['max-lease-time']

        for change in change_set:
            if not change["xpath"].startswith(self.namespace):
                continue
            if not 'operation' in change:
                self.logger.error('Malformed change set')
                continue

            self.logger.debug('Operation -->> %s' % change['operation'])
            self.logger.debug('XPath -->> %s' % change['xpath'])

            if change['operation'] == 'MODIFIED':
                if "enabled" in change["xpath"]:
                    self.logger.debug('Modified  - %s: %s -> %s' % (change['xpath'], change['old_value'], change['new_value']))
                    dhcp_state = change['new_value']
                if "dhcp-range" in change["xpath"]:
                    self.logger.debug('Modified  - %s: %s -> %s' % (change['xpath'], change['old_value'], change['new_value']))
                    dhcp_range = change['new_value']
                if "dhcp-bind-interface" in change["xpath"]:
                    self.logger.debug('Modified  - %s: %s -> %s' % (change['xpath'], change['old_value'], change['new_value']))
                    dhcp_int = change['new_value']
                if "router" in change["xpath"]:
                    self.logger.debug('Modified  - %s: %s -> %s' % (change['xpath'], change['old_value'], change['new_value']))
                    router = change['new_value']
                if "domain-name-servers" in change["xpath"]:
                    self.logger.debug('Modified  - %s: %s -> %s' % (change['xpath'], change['old_value'], change['new_value']))
                    dns = change['new_value']
                if "default-lease-time" in change["xpath"]:
                    self.logger.debug('Modified  - %s: %s -> %s' % (change['xpath'], change['old_value'], change['new_value']))
                    dlt = change['new_value']
                if "max-lease-time" in change["xpath"]:
                    self.logger.debug('Modified  - %s: %s -> %s' % (change['xpath'], change['old_value'], change['new_value']))
                    mlt = change['new_value']

            elif change['operation'] == 'MOVED':
                self.logger.debug('Moved  - %s' % change['xpath'])
            elif change['operation'] == 'DELETED':
                router = None
                dns = None
                dlt = None
                mlt = None
                self.logger.debug('Deleted  - %s' % change['xpath'])
            elif change['operation'] == 'CREATED':
                if "enabled" in change["xpath"]:
                    self.logger.debug('Created  - %s' % change['new_value'])
                    dhcp_state = change['new_value']
                if "dhcp-range" in change["xpath"]:
                    self.logger.debug('Created  - %s' % change['new_value'])
                    dhcp_range = change['new_value']
                if "dhcp-bind-interface" in change["xpath"]:
                    self.logger.debug('Created  - %s' % change['new_value'])
                    dhcp_int = change['new_value']
                if "router" in change["xpath"]:
                    self.logger.debug('Created  - %s' % change['new_value'])
                    router = change['new_value']
                if "domain-name-servers" in change["xpath"]:
                    self.logger.debug('Created  - %s' % change['new_value'])
                    dns = change['new_value']
                if "default-lease-time" in change["xpath"]:
                    self.logger.debug('Created  - %s' % change['new_value'])
                    dlt = change['new_value']
                if "max-lease-time" in change["xpath"]:
                    self.logger.debug('Created  - %s' % change['new_value'])
                    mlt = change['new_value']

            else:
                self.logger.error('Unknown operation \'%s\'' % change['operation'])

        if dhcp_state and 'true' in str(dhcp_state).lower():
            if self.isIPv6(dhcp_range):
                dhcp_service = "isc-dhcp-server6.service"
                addr = self.get_system_bind_iface6_address(dhcp_int)
                netmask = self.get_system_bind_iface6_netmask(dhcp_int)
                res = self.apply_dhcp6server2system(addr, netmask, dhcp_range, router, dns, dlt, mlt, dhcp_int)
            else:
                dhcp_service = "isc-dhcp-server.service"
                addr = self.get_system_bind_iface_address(dhcp_int)
                netmask = self.get_system_bind_iface_netmask(dhcp_int)
                res = self.apply_dhcpserver2system(addr, netmask, dhcp_range, router, dns, dlt, mlt)

            if res is False:
                self.logger.error('Failed creating DHCP server configuration file')
                return True
            self.start_dhcp_service(dhcp_service)
        else:
            self.stop_dhcp_service("isc-dhcp-server.service")
            self.stop_dhcp_service("isc-dhcp-server6.service")

        return True

    def apply_config(self, change_set=None, apply_startup=False):
        ret = False
        if apply_startup:
            ret = self.apply_startup()
        else:
            ret = self.apply_running(change_set)

        return ret

    def set_state(self, state):
        self.set_item("{}/enabled".format(self.namespace), state)

    def get_state(self):
        return self.get_item("{}/enabled".format(self.namespace))

    def unset_state(self):
        self.delete_item("{}/enabled".format(self.namespace))

    def set_dhcp_range(self, dhcp_range):
        self.set_item("{}/dhcp-range".format(self.namespace), dhcp_range)

    def get_dhcp_range(self):
        return self.get_item("{}/dhcp-range".format(self.namespace))

    def unset_dhcp_range(self):
        self.delete_item("{}/dhcp-range".format(self.namespace))

    def set_dhcp_int(self, dhcp_int):
        self.set_item("{}/dhcp-bind-interface".format(self.namespace), dhcp_int)

    def get_dhcp_int(self):
        return self.get_item("{}/dhcp-bind-interface".format(self.namespace))

    def unset_dhcp_int(self):
        self.delete_item("{}/dhcp-bind-interface".format(self.namespace))

    def set_dhcp_router(self, router):
        self.set_item("{}/router".format(self.namespace), router)

    def get_dhcp_router(self):
        return self.get_item("{}/router".format(self.namespace))

    def unset_dhcp_router(self):
        self.delete_item("{}/router".format(self.namespace))

    def set_dhcp_domain_name_servers(self, dns):
        self.set_item("{}/domain-name-servers".format(self.namespace), dns)

    def get_dhcp_domain_name_servers(self):
        return self.get_item("{}/domain-name-servers".format(self.namespace))

    def unset_dhcp_domain_name_servers(self):
        self.delete_item("{}/domain-name-servers".format(self.namespace))

    def set_dhcp_default_lease_time(self, dlt):
        self.set_item("{}/default-lease-time".format(self.namespace), dlt)

    def get_dhcp_default_lease_time(self):
        return self.get_item("{}/default-lease-time".format(self.namespace))

    def unset_dhcp_default_lease_time(self):
        self.delete_item("{}/default-lease-time".format(self.namespace))

    def set_dhcp_max_lease_time(self, mlt):
        self.set_item("{}/max-lease-time".format(self.namespace), mlt)

    def get_dhcp_max_lease_time(self):
        return self.get_item("{}/max-lease-time".format(self.namespace))

    def unset_dhcp_max_lease_time(self):
        self.delete_item("{}/max-lease-time".format(self.namespace))
